# Pre-Requisites --> 
# Python
#     PIP Gast, Tensorflow, matplotlib
# vMicro   
#     Board Package for ESP32, Tensorflow_ESP32 Library, ESP32_FTPClient Library

# FTP Server to Receive Files (parsed by this script)

# Full Cloud Build 
# Azure Account for Dev Ops, and configured as a Build Pipeline Agent

# 
# https://colab.research.google.com/github/tensorflow/tensorflow/blob/master/tensorflow/lite/micro/examples/hello_world/train/train_hello_world_model.ipynb#scrollTo=aCZBFzjClURz

# Define paths to model files
import os
MODELS_DIR = 'models/'
if not os.path.exists(MODELS_DIR):
    os.mkdir(MODELS_DIR)
MODEL_TF = MODELS_DIR + 'model.pb'
MODEL_NO_QUANT_TFLITE = MODELS_DIR + 'model_no_quant.tflite'
MODEL_TFLITE = MODELS_DIR + 'model.tflite'
MODEL_TFLITE_MICRO = MODELS_DIR + 'model.cc'

# Find Path Script is executing from
import sys
basePath = sys.path[0].replace("\\","/")


# Set a "seed" value, so we get the same random numbers each time we run this
# notebook for reproducible results.
# Numpy is a math library
import numpy as np
np.random.seed(1) # numpy seed
# TensorFlow is an open source machine learning library
import tensorflow as tf
tf.random.set_seed(1) # tensorflow global random seed

# Keras is TensorFlow's high-level API for deep learning
from tensorflow import keras
# Matplotlib is a graphing library
import matplotlib.pyplot as plt
# Math is Python's math library
import math


####################################################################################################
#    CHANGE TO READ IN COMPILED CSV FILES FROM FTP SERVER ROOT
####################################################################################################
# # # # Number of sample datapoints
# # # SAMPLES = 1000

# # # # Generate a uniformly distributed set of random numbers in the range from
# # # # 0 to 2π, which covers a complete sine wave oscillation
# # # x_values = np.random.uniform(
    # # # low=0, high=0.1, size=SAMPLES).astype(np.float32)

# # # # Shuffle the values to guarantee they're not in order
# # # np.random.shuffle(x_values)

# # # # Calculate the corresponding sine values
# # # y_values = np.sin(x_values).astype(np.float32)

# # # # Plot our data. The 'b.' argument tells the library to print blue dots.
# # # plt.plot(x_values, y_values, 'b.')
# # # plt.show()


# # # # Any Randomisation would naturally appear in samples (as it does from board)
# # # # Add a small random number to each y value
# # # # y_values += 0.1 * np.random.randn(*y_values.shape)

# # # # Plot our data
# # # plt.plot(x_values, y_values, 'b.')
# # # plt.show()


with open(basePath + "/FTPRoot/devicedata.csv", newline='') as csvfile:
    csvcontent = csvfile.readlines()
csvcontent = [x.strip() for x in csvcontent]

samplesCounted = 0
xVals = []


yVals = []

for row in csvcontent:
    xVals.append(row.split(',')[0])
    yVals.append(row.split(',')[1])
    samplesCounted += 1

SAMPLES = samplesCounted

x_values = np.array(xVals, dtype='f')
y_values = np.array(yVals, dtype='f')

# Shuffle the values to guarantee they're not in order
# np.random.shuffle(x_values)

# Plot our data. The 'b.' argument tells the library to print blue dots.
plt.plot(x_values, y_values, 'b.')
plt.show()

####################################################################################################

# We'll use 60% of our data for training and 20% for testing. The remaining 20%
# will be used for validation. Calculate the indices of each section.
TRAIN_SPLIT =  int(0.6 * SAMPLES)
TEST_SPLIT = int(0.2 * SAMPLES + TRAIN_SPLIT)

# Use np.split to chop our data into three parts.
# The second argument to np.split is an array of indices where the data will be
# split. We provide two indices, so the data will be divided into three chunks.
x_train, x_test, x_validate = np.split(x_values, [TRAIN_SPLIT, TEST_SPLIT])
y_train, y_test, y_validate = np.split(y_values, [TRAIN_SPLIT, TEST_SPLIT])

# Double check that our splits add up correctly
assert (x_train.size + x_validate.size + x_test.size) ==  SAMPLES

# Plot the data in each partition in different colors:
plt.plot(x_train, y_train, 'b.', label="Train")
plt.plot(x_test, y_test, 'r.', label="Test")
plt.plot(x_validate, y_validate, 'y.', label="Validate")
plt.legend()
plt.show()

# We'll use Keras to create a simple model architecture
model_2 = tf.keras.Sequential()

# First layer takes a scalar input and feeds it through 16 "neurons". The
# neurons decide whether to activate based on the 'relu' activation function.
model_2.add(keras.layers.Dense(16, activation='relu', input_shape=(1,)))

# The new second layer may help the network learn more complex representations
model_2.add(keras.layers.Dense(16, activation='relu'))

# Final layer is a single neuron, since we want to output a single value
model_2.add(keras.layers.Dense(1))

# Compile the model using a standard optimizer and loss function for regression
model_2.compile(optimizer='adam', loss='mse', metrics=['mae'])

history_2 = model_2.fit(x_train, y_train, epochs=500, batch_size=64,
                    validation_data=(x_validate, y_validate))

# Draw a graph of the loss, which is the distance between
# the predicted and actual values during training and validation.
loss = history_2.history['loss']
val_loss = history_2.history['val_loss']

epochs = range(1, len(loss) + 1)

# Exclude the first few epochs so the graph is easier to read
SKIP = 100

plt.figure(figsize=(10, 4))
plt.subplot(1, 2, 1)

plt.plot(epochs[SKIP:], loss[SKIP:], 'g.', label='Training loss')
plt.plot(epochs[SKIP:], val_loss[SKIP:], 'b.', label='Validation loss')
plt.title('Training and validation loss')
plt.xlabel('Epochs')
plt.ylabel('Loss')
plt.legend()

plt.subplot(1, 2, 2)

# Draw a graph of mean absolute error, which is another way of
# measuring the amount of error in the prediction.
mae = history_2.history['mae']
val_mae = history_2.history['val_mae']

plt.plot(epochs[SKIP:], mae[SKIP:], 'g.', label='Training MAE')
plt.plot(epochs[SKIP:], val_mae[SKIP:], 'b.', label='Validation MAE')
plt.title('Training and validation mean absolute error')
plt.xlabel('Epochs')
plt.ylabel('MAE')
plt.legend()

plt.tight_layout()

# Calculate and print the loss on our test dataset
loss = model_2.evaluate(x_test, y_test)

# Make predictions based on our test dataset
predictions = model_2.predict(x_test)

# Graph the predictions against the actual values
plt.clf()
plt.title('Comparison of predictions and actual values')
plt.plot(x_test, y_test, 'b.', label='Actual')
plt.plot(x_test, predictions, 'r.', label='Predicted')
plt.legend()
plt.show()

# Convert the model to the TensorFlow Lite format without quantization
converter = tf.lite.TFLiteConverter.from_keras_model(model_2)
model_no_quant_tflite = converter.convert()

# # Save the model to disk
open(MODEL_NO_QUANT_TFLITE, "wb").write(model_no_quant_tflite)

# Convert the model to the TensorFlow Lite format with quantization
# Scale data set to 500 points 
def representative_dataset():
  xrange = x_train.size
  if (x_train.size > 500):
    xrange = 500
  for i in range(xrange):
    yield([x_train[i].reshape(1, 1)])
# Set the optimization flag.
converter.optimizations = [tf.lite.Optimize.DEFAULT]
# Enforce full-int8 quantization (except inputs/outputs which are always float)
converter.target_spec.supported_ops = [tf.lite.OpsSet.TFLITE_BUILTINS_INT8]
# Provide a representative dataset to ensure we quantize correctly.
converter.representative_dataset = representative_dataset
model_tflite = converter.convert()

# Save the model to disk
open(MODEL_TFLITE, "wb").write(model_tflite)


import os
model_no_quant_size = os.path.getsize(MODEL_NO_QUANT_TFLITE)
print("Model is %d bytes" % model_no_quant_size)
# model_size = os.path.getsize(MODEL_TFLITE)
# print("Quantized model is %d bytes" % model_size)
# difference = model_no_quant_size - model_size
# print("Difference is %d bytes" % difference)

# Instantiate an interpreter for each model
model_no_quant = tf.lite.Interpreter(MODEL_NO_QUANT_TFLITE)

# Allocate memory for each model
model_no_quant.allocate_tensors()

# Get the input and output tensors so we can feed in values and get the results
model_no_quant_input = model_no_quant.tensor(model_no_quant.get_input_details()[0]["index"])
model_no_quant_output = model_no_quant.tensor(model_no_quant.get_output_details()[0]["index"])


# Create arrays to store the results
model_no_quant_predictions = np.empty(x_test.size)

# Run each model's interpreter for each value and store the results in arrays
for i in range(x_test.size):
  model_no_quant_input().fill(x_test[i])
  model_no_quant.invoke()
  model_no_quant_predictions[i] = model_no_quant_output()[0]

# See how they line up with the data
plt.clf()
plt.title('Comparison of various models against actual values')
plt.plot(x_test, y_test, 'bo', label='Actual values')
plt.plot(x_test, model_no_quant_predictions, 'bx', label='Lite predictions')
plt.legend()
plt.show()


# Additional Wrapping for Bash on Windows (from MSYS2 Installation)

# Exec XXD to build our C++ Code for us from the model
os.system("F:\\Espressif\\msys32\\usr\\bin\\bash.exe --login -c \"xxd -i " + basePath + "/" + MODEL_NO_QUANT_TFLITE + " > " + basePath + "/" + MODEL_TFLITE_MICRO + "\"")

# Now update our sketch code from the data generated in the model.....
REPLACE_TEXT = (basePath + "/" + MODEL_NO_QUANT_TFLITE).replace('/', '_').replace('.', '_').replace(':', '_')

with open(basePath + "/" + MODEL_TFLITE_MICRO) as f:
    content = f.readlines()
# you may also want to remove whitespace characters like `\n` at the end of each line
content = [x.strip() for x in content] 
content = [x.replace(REPLACE_TEXT,"g_sine_model_data") for x in content]
content = [x.replace("g_sine_model_data[] ","g_sine_model_data[] DATA_ALIGN_ATTRIBUTE ") for x in content]

# Can now Edit the File in the Sketch
with open(basePath + "/ESP32EdgeDevice/sine_model_data.cpp", 'r') as f:
    oldcontent = f.readlines()
with open(basePath + "/ESP32EdgeDevice/sine_model_data.cpp", 'w') as f:	
    for ln in oldcontent:
	    if "g_sine_model_data[]" in ln:
		    break
	    else:
		    f.write(ln)
    for ln in content:
        newln = ln.replace("unsigned int", "const int").replace("unsigned char","const unsigned char")
        f.write(newln + "\n")

		
# Now fall back to vMicro Azure Build Script // VS Build Hooks

